#pragma once

#include <cstdint>
#include <cmath>
#include "Platform.h"

namespace GenTLConsumerImplHelper
{
    template<typename ParameterPtr, typename ValueType>
    inline ValueType GetParameterValue(ParameterPtr parameter)
    {
        if (!parameter.IsValid())
        {
            throw RUNTIME_EXCEPTION("Uninitialized parameter pointer.");
        }
        try
        {
            return parameter->GetValue();
        }
        catch (const GenICam::GenericException& e)
        {
            CIH_RETHROW_GENICAM_EXCEPTION(e, "Failed to read parameter %s: %s",
                parameter->GetNode()->GetName().c_str(),
                e.GetDescription());
        }
    }

    template<typename ParameterPtr, typename ValueType>
    inline void SetParameterValue(ParameterPtr parameter, ValueType value)
    {
        if (!parameter.IsValid())
        {
            throw RUNTIME_EXCEPTION("Uninitialized parameter pointer.");
        }
        try
        {
            parameter->SetValue(value);
        }
        catch (const GenICam::GenericException& e)
        {
            CIH_RETHROW_GENICAM_EXCEPTION(e, "Failed to set parameter %s: %s",
                parameter->GetNode()->GetName().c_str(),
                e.GetDescription());
        }
    }

    template<> inline std::string GetParameterValue< GenApi::CValuePtr, std::string>(GenApi::CValuePtr parameter)
    {
        if (!parameter.IsValid())
        {
            throw RUNTIME_EXCEPTION("Uninitialized parameter pointer.");
        }
        try
        {
            return parameter->ToString().c_str();
        }
        catch (const GenICam::GenericException& e)
        {
            CIH_RETHROW_GENICAM_EXCEPTION(e, "Failed to read parameter %s: %s",
                parameter->GetNode()->GetName().c_str(),
                e.GetDescription());
        }
    }

    template<> inline void SetParameterValue< GenApi::CValuePtr, const std::string&>(GenApi::CValuePtr parameter, const std::string& value)
    {
        if (!parameter.IsValid())
        {
            throw RUNTIME_EXCEPTION("Uninitialized parameter pointer.");
        }
        try
        {
            parameter->FromString(value.c_str());
        }
        catch (const GenICam::GenericException& e)
        {
            CIH_RETHROW_GENICAM_EXCEPTION(e, "Failed to set parameter %s: %s",
                parameter->GetNode()->GetName().c_str(),
                e.GetDescription());
        }
    }

    template<> inline void SetParameterValue< GenApi::CValuePtr, const char*>(GenApi::CValuePtr parameter, const char* value)
    {
        if (!parameter.IsValid())
        {
            throw RUNTIME_EXCEPTION("Uninitialized parameter pointer.");
        }
        try
        {
            parameter->FromString(value);
        }
        catch (const GenICam::GenericException& e)
        {
            CIH_RETHROW_GENICAM_EXCEPTION(e, "Failed to set parameter %s: %s",
                parameter->GetNode()->GetName().c_str(),
                e.GetDescription());
        }
    }

    template<> inline void SetParameterValue<GenApi::CIntegerPtr, int64_t>(GenApi::CIntegerPtr parameter, int64_t value)
    {
        if (!parameter.IsValid())
        {
            throw RUNTIME_EXCEPTION("Uninitialized parameter pointer.");
        }
        try
        {
            const auto min = parameter->GetMin();
            const auto max = parameter->GetMax();
            const auto inc = parameter->GetInc();

            if (value < min)
            {
                value = min;
            }
            else if (value > max)
            {
                value = max;
            }
            else if (inc >= 1)
            {
                value = min + (((value - min) / inc) * inc);
            }
            parameter->SetValue(value);
        }
        catch (const GenICam::GenericException& e)
        {
            CIH_RETHROW_GENICAM_EXCEPTION(e, "Failed to set parameter %s: %s",
                parameter->GetNode()->GetName().c_str(),
                e.GetDescription());
        }
    }

    template<> inline void SetParameterValue<GenApi::CFloatPtr, double>(GenApi::CFloatPtr parameter, double value)
    {
        if (!parameter.IsValid())
        {
            throw RUNTIME_EXCEPTION("Uninitialized parameter pointer.");
        }
        try
        {
            const auto min = parameter->GetMin();
            const auto max = parameter->GetMax();

            if (value < min)
            {
                value = min;
            }
            else
            {
                if (parameter->GetIncMode() == GenApi::fixedIncrement)
                {
                    const auto inc = parameter->GetInc();
                    value = min + (std::round((value - min) / inc) * inc);
                }
                if (value > max)
                {
                    value = max;
                }
            }
            parameter->SetValue(value);
        }
        catch (const GenICam::GenericException& e)
        {
            CIH_RETHROW_GENICAM_EXCEPTION(e, "Failed to set parameter %s: %s",
                parameter->GetNode()->GetName().c_str(),
                e.GetDescription());
        }
    }

    void inline ExecuteCommand(GenApi::CCommandPtr command)
    {
        if (!command.IsValid())
        {
            throw RUNTIME_EXCEPTION("Uninitialized parameter pointer.");
        }
        try
        {
            command->Execute();
        }
        catch (GenICam::GenericException& e)
        {
            CIH_RETHROW_GENICAM_EXCEPTION(e, "Failed to execute command %s: %s",
                command->GetNode()->GetName().c_str(),
                e.GetDescription());
        }
    }
}
